/*
    Copyright (c) 2009 by Joseph Devietti (devietti@cs.washington.edu).

    This file is part of my DMP ASPLOS 2009 Presentation (DAP).

    DAP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DAP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DAP.  If not, see <http://www.gnu.org/licenses/>.
*/ 

﻿package  {
	
	import com.boostworthy.geom.Path;
	import com.dncompute.graphics.*;
	import flash.display.*;
	import flash.events.*;
	import flash.text.*;
	import flash.geom.Point;
	import flash.ui.*;
	import com.boostworthy.animation.management.*;
	import com.boostworthy.animation.easing.*;
	import com.boostworthy.animation.sequence.*;
	import com.boostworthy.animation.sequence.tweens.*;
	import com.boostworthy.animation.rendering.*;
	import com.boostworthy.core.Global;
	
	/**
	 * Animation for DMP serial explanation.
	 * @author Joe Devietti
	 */
	public class TM extends Sprite {
		
	public function TM() {
		if (stage) init();
		else addEventListener(Event.ADDED_TO_STAGE, init);
	}
	
	private const g_timeline:Timeline = new Timeline( RenderMethod.ENTER_FRAME );
	
	private const XACT_ABORTED_ALPHA:Number = 0.2;
	
	private function init(e:Event = null):void {
		removeEventListener(Event.ADDED_TO_STAGE, init);
		// entry point
		
		SceneManager.setup( g_timeline, stage );


// SHOW CODE, CPUS AND TOKEN
	// local iteration vars
	var i:int = 0;
	var s:Sprite = null; var t:Sprite = null;
	var w:Object = null;

	var sp_code_cpus:Array = [];
	for ( i = 0; i < ExecLayout.NUM_CPUS; i++ ) {
		// draw codebox
		s = MyShapes.CodeBox( ExecLayout.INSNS[i] );
		s.x = ExecLayout.COLUMN_START[i] + ExecLayout.CODE_GAP;
		s.y = ExecLayout.CODE_TOP;
		addChild( s );
		sp_code_cpus.push( s );
		
		// draw cpu
		s = MyShapes.Cpu( "p"+i );
		s.x = ExecLayout.COLUMN_START[i] + ExecLayout.CPU_GAP;
		s.y = ExecLayout.CPU_TOP;
		addChild( s );
		sp_code_cpus.push( s );
	}
	
	// draw token at p0
	var sp_token:Sprite = MyShapes.Token();
	sp_token.x = ExecLayout.tokenAt(0).x;
	sp_token.y = ExecLayout.tokenAt(0).y;
	ExecLayout.g_current_token_cpu = 0;
	addChild( sp_token );
	
	// time arrows
	var sp_time_arrows:Array = [];
	for ( i = 1; i <= 2; i++ ) {
		s = MyShapes.TimeArrow();
		s.x = ExecLayout.COLUMN_MIDPOINT_START[i] + (s.width/2);
		s.y = ExecLayout.INSN_TOP;
		addChild( s );
		sp_time_arrows.push( s );
	}


// ALL PROCS EXEC QUANTA SPECULATIVELY
var sp_quanta:Array = [];
var sp_insns:Array = [];
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// local iteration vars
	var i:int = 0; var j:int = 0;
	var s:Sprite = null; var t:Sprite = null;
	
	for ( i = 0; i < ExecLayout.NUM_CPUS; i++ ) {
		
		// draw xact
		var xact:Sprite = MyShapes.Xact(2);
		xact.x = ExecLayout.COLUMN_START[i] + ExecLayout.XACT_GAP;
		xact.y = ExecLayout.XACT_TOP;
		xact.alpha = 0.0;
		xact.name = "xact"+i+"a";
		addChild( xact );
		sp_quanta.push( xact );
		g_timeline.addTween( new Tween( xact, "alpha", 1.0, Anim.Start(), Anim.StopNoIncr(Anim.INSN_APPEAR), Transitions.LINEAR) );		
		
		// draw insns
		for ( j = 0; j < ExecLayout.INSNS_PER_CPU; j++ ) {
			s = MyShapes.Insn( ExecLayout.INSNS[i][j], MyShapes.INSN_COLORS[i] );
			s.name = (("i"+i)+j)+"a";
			sp_insns.push( s );
			
			if ( i == 2 ) { // we want to extend just the xact box later, which is easier if it doesn't have children
				addChild( s );
				s.x = xact.localToGlobal( new Point(MyShapes.XACT_XOFFSET,0) ).x;
				s.y = xact.localToGlobal( new Point(0,MyShapes.XACT_YOFFSET + (j*MyShapes.INSN_YOFFSET)) ).y;
			} else {
				xact.addChild( s ); // add to xact to make it easier to dim whole group
				s.x = MyShapes.XACT_XOFFSET;
				s.y = MyShapes.XACT_YOFFSET + (j*MyShapes.INSN_YOFFSET);
			}
			s.alpha = 0.0;
			g_timeline.addTween( new Tween( s, "alpha", 1.0, Anim.Start(), Anim.StopNoIncr(Anim.INSN_APPEAR), Transitions.LINEAR) );
		}
	}
	Anim.StopIncr(Anim.INSN_APPEAR);
},
function():void {
	for each ( var s:Sprite in sp_quanta ) { removeChild(s); }
	sp_quanta = [];
	sp_insns = [];
},
Anim.INSN_APPEAR);


// COMMIT XACT0
var sp_commits:Array = [];
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	
	var commit:Sprite = MyShapes.XactCommit();
	commit.x = ExecLayout.COLUMN_START[0] + ExecLayout.XACT_GAP;
	
	var xact:Sprite = ExecLayout.findByName( sp_quanta, "xact0a" );
	commit.y = xact.y + xact.height - MyShapes.XACT_LINE_THICKNESS;
	
	//commit.alpha = 0.0;
	commit.name = "commit0";
	addChild( commit );
	sp_commits.push( commit );
	w = Wiper.wipe( commit, Wiper.WIPE_RIGHT, true );
	g_timeline.addTween( new Tween(w.mask,"x",w.x,Anim.Start(),Anim.StopIncr(Anim.XACT_COMMIT),Transitions.LINEAR) );
},
function():void {
	for each ( var s:Sprite in sp_commits ) { removeChild(s); }
	sp_commits = [];
},
Anim.XACT_COMMIT);


// CONFLICT BETWEEN XACT0 AND XACT1
var sp_conflict_arrow01:Sprite = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// draw conflict arrow
	var from:Sprite = ExecLayout.findByName( sp_insns, "i00a" );
	var fromP:Point = from.localToGlobal( new Point(from.width,(from.height/2)) );
	var to:Sprite = ExecLayout.findByName( sp_insns, "i11a" );
	var toP:Point = to.localToGlobal( new Point(0,(to.height/2)) );
	sp_conflict_arrow01 = MyShapes.ConflictArrow( fromP.x, fromP.y, toP.x, toP.y, "CONFLICT" );
	addChild( sp_conflict_arrow01 );
	
	g_timeline.addTween( ExecLayout.pathWipeTLtoBR(sp_conflict_arrow01, Anim.COMM_ARROW_WIPE) );
},
function():void {
	removeChild( sp_conflict_arrow01 );
	sp_conflict_arrow01 = null;
},
Anim.COMM_ARROW_WIPE);


// REEXECUTE XACT1
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// local iteration vars
	var i:int = 0; var j:int = 0;
	var s:Sprite = null; var aborted:Sprite = null;
	
	// get rid of conflict arrow
	g_timeline.addTween( ExecLayout.fadeToNoIncr(sp_conflict_arrow01,0.0) );
	// dim aborted xact
	aborted = ExecLayout.findByName( sp_quanta, "xact1a" );
	g_timeline.addTween( new Tween( aborted, "alpha", XACT_ABORTED_ALPHA, 
		Anim.Start(), Anim.StopIncr(Anim.INSN_APPEAR), Transitions.LINEAR) );
		
	// draw new xact
	i = 1;
	var xact:Sprite = MyShapes.Xact(2);
	xact.x = aborted.x;
	
	//var y0:Number = ExecLayout.XACT_TOP + MyShapes.xactHeight(2) + MyShapes.XACT_YOFFSET;
	var y1:Number = aborted.y + aborted.height + MyShapes.XACT_YOFFSET;
	//trace( y0 + "<- old   new ->" + y1 ); // old=222 new=225, but i like new spacing better
	xact.y = y1;
	
	xact.alpha = 0.0;
	xact.name = "xact"+i+"b";
	addChild( xact );
	sp_quanta.push( xact );
	g_timeline.addTween( new Tween( xact, "alpha", 1.0, 
		Anim.Start(), Anim.StopNoIncr(Anim.INSN_APPEAR), Transitions.LINEAR) );		
		
	// draw insns
	for ( j = 0; j < ExecLayout.INSNS_PER_CPU; j++ ) {
		s = MyShapes.Insn( ExecLayout.INSNS[i][j], MyShapes.INSN_COLORS[i] );
		s.name = (("i"+i)+j)+"b";
		sp_insns.push( s );
		xact.addChild( s );
		s.x = MyShapes.XACT_XOFFSET;
		s.y = MyShapes.XACT_YOFFSET + (j*MyShapes.INSN_YOFFSET);
		s.alpha = 0.0;
		g_timeline.addTween( new Tween( s, "alpha", 1.0, Anim.Start(), Anim.StopNoIncr(Anim.INSN_APPEAR), Transitions.LINEAR) );
	}
	Anim.StopIncr(Anim.INSN_APPEAR);
},
function():void {
	sp_quanta = ExecLayout.removeByName( ["xact1b"], sp_quanta, removeChild );
	sp_insns = ExecLayout.removeByName( ["i10b","i11b"], sp_insns, function(s:Sprite):void{} );
},
Anim.INSN_APPEAR*2);


// PASS TOKEN, COMMIT XACT1
SceneManager.addScene(
function(start_frame:uint, end_frame:uint):void {
	
	// pass token to p1
	g_timeline.addTween( ExecLayout.moveTokenTo(sp_token,1) );
	
	// commit xact1
	var commit:Sprite = MyShapes.XactCommit();
	commit.x = ExecLayout.COLUMN_START[1] + ExecLayout.XACT_GAP;
	
	var xact:Sprite = ExecLayout.findByName( sp_quanta, "xact1b" );
	commit.y = xact.y + xact.height - MyShapes.XACT_LINE_THICKNESS;
	
	commit.name = "commit1";
	addChild( commit );
	sp_commits.push( commit );
	w = Wiper.wipe( commit, Wiper.WIPE_RIGHT, true );
	g_timeline.addTween( new Tween(w.mask,"x",w.x,Anim.Start(),Anim.StopIncr(Anim.XACT_COMMIT),Transitions.LINEAR) );
},
function():void {
	sp_commits = ExecLayout.removeByName( ["commit1"], sp_commits, removeChild );
},
Anim.MOVE_TOKEN + Anim.XACT_COMMIT);


// SHOW FALSE CONFLICT WITH XACT2
var sp_conflict_arrow12:Sprite = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// draw conflict arrow
	var from:Sprite = ExecLayout.findByName( sp_insns, "i10b" );
	var fromP:Point = from.localToGlobal( new Point(from.width,(from.height/2)) );
	var to:Sprite = ExecLayout.findByName( sp_insns, "i21a" );
	var toP:Point = to.localToGlobal( new Point(0,(to.height/2)) );
	sp_conflict_arrow12 = MyShapes.ConflictArrow( fromP.x, fromP.y, toP.x, toP.y, "CONFLICT?" );
	addChild( sp_conflict_arrow12 );

	g_timeline.addTween( ExecLayout.pathWipeBLtoTR(sp_conflict_arrow12, Anim.COMM_ARROW_WIPE) );
},
function():void {
	removeChild( sp_conflict_arrow12 );
	sp_conflict_arrow12 = null;
},
Anim.COMM_ARROW_WIPE);


// EXTEND XACT2, PASS TOKEN AND COMMIT
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {

	// remove false conflict
	g_timeline.addTween( ExecLayout.fadeToIncr(sp_conflict_arrow12,0.0) );
	
	var xact2:Sprite = ExecLayout.findByName( sp_quanta, "xact2a" );
	
	// draw new xact2
	i = 2;
	var new_xact:Sprite = MyShapes.Xact(4.65);
	var new_xact_height:int = new_xact.height; // NB: need to save this here as adding mask changes height!
	new_xact.x = xact2.x;
	new_xact.y = xact2.y;
	new_xact.name = "xact"+i+"b";
	addChildAt( new_xact, getChildIndex(sp_conflict_arrow12)-1 ); // draw below 1->2 conflict arrow
	sp_quanta.push( new_xact );
	
	// wipe in larger xact
	w = Wiper.wipe( new_xact, Wiper.WIPE_DOWN, true );
	g_timeline.addTween( new Tween(w.mask,"y",w.y,Anim.Start(),Anim.StopNoIncr(Anim.XACT_COMMIT),Transitions.LINEAR) );
	
	// wipe out old xact to avoid overlap
	var stop:int = Anim.Start() + ((xact2.height / new_xact_height) * Anim.XACT_COMMIT);
	w = Wiper.wipe( xact2, Wiper.WIPE_DOWN, false );
	g_timeline.addTween( new Tween(w.mask,"y",w.y,Anim.Start(),stop,Transitions.LINEAR) );
	Anim.StopIncr(Anim.XACT_COMMIT);
	
	// pass token to p2
	g_timeline.addTween( ExecLayout.moveTokenTo(sp_token,2) );
	
	// commit xact2
	var commit:Sprite = MyShapes.XactCommit();
	commit.x = ExecLayout.COLUMN_START[2] + ExecLayout.XACT_GAP;
	
	commit.y = xact2.y + new_xact_height - MyShapes.XACT_LINE_THICKNESS;
	
	commit.name = "commit2";
	addChild( commit );
	sp_commits.push( commit );
	var ww:Object = Wiper.wipe( commit, Wiper.WIPE_RIGHT, true );
	g_timeline.addTween( new Tween(ww.mask,"x",ww.x,Anim.Start(),Anim.StopIncr(Anim.XACT_COMMIT),Transitions.LINEAR) );
},
function():void {
	sp_commits = ExecLayout.removeByName( ["commit2"], sp_commits, removeChild );
	sp_quanta = ExecLayout.removeByName( ["xact2b"], sp_quanta, removeChild );
},
Anim.FADE + (2*Anim.XACT_COMMIT) + Anim.MOVE_TOKEN);


// ADD MEMORY RENAMING NOTE
var sp_memrenaming_text:DisplayObject = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	var txt:TextField = new UnselectableTextField();
	txt.htmlText = "ordering + isolation = memory renaming";
	txt.autoSize = TextFieldAutoSize.LEFT;
	txt.setTextFormat( MyShapes.getLabelTextFormat() );
	addChild( txt );
	txt.x = (ExecLayout.STAGE_WIDTH - txt.width) / 2;
	txt.y = 330;
	txt.alpha = 0;
	
	sp_memrenaming_text = txt;
	g_timeline.addTween( ExecLayout.fadeToIncr(txt,1.0) );
},
function():void {
	removeChild( sp_memrenaming_text );
	sp_memrenaming_text = null;
},
Anim.FADE);


// SHOW COMMUNICATION ARROWS
var sp_comm_arrows:Array = [];
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// dim code and CPU's and renaming note
	for each ( var s:Sprite in sp_code_cpus.concat(sp_token,sp_time_arrows) ) {
		g_timeline.addTween( ExecLayout.fadeToNoIncr(s,ExecLayout.ALPHA_DIM) );
	}
	g_timeline.addTween( ExecLayout.fadeToNoIncr(sp_memrenaming_text,0.0) );	
	Anim.StopIncr(Anim.FADE);
	
	// wipe in communication arrows
	var from:Sprite = ExecLayout.findByName( sp_insns, "i00a" );
	var fromP:Point = from.parent.localToGlobal( new Point(from.x+from.width,from.y+(from.height/2)) );
	var to:Sprite = ExecLayout.findByName( sp_insns, "i11b" );
	var toP:Point = to.parent.localToGlobal( new Point(to.x,to.y+(to.height/2)) );

	var arrow:Sprite = new TextLineArrow( fromP.x-ExecLayout.COMM_ARROW_XOFFSET, fromP.y, 
		toP.x+ExecLayout.COMM_ARROW_XOFFSET, toP.y, "RAW", ExecLayout.COMM_ARROW_COLOR );
	arrow.name = "raw";
	addChild( arrow );
	sp_comm_arrows.push( arrow );
	
	g_timeline.addTween( ExecLayout.pathWipeTLtoBR(arrow, Anim.COMM_ARROW_WIPE) );
	
	// wipe in communication arrows
	from = ExecLayout.findByName( sp_insns, "i10b" );
	fromP = from.parent.localToGlobal( new Point(from.x+from.width,from.y+(from.height/2)) );
	to = ExecLayout.findByName( sp_insns, "i21a" );
	toP = to.parent.localToGlobal( new Point(to.x,to.y+(to.height/2)) );

	arrow = new TextLineArrow( fromP.x-ExecLayout.COMM_ARROW_XOFFSET, fromP.y, 
		toP.x+ExecLayout.COMM_ARROW_XOFFSET, toP.y, "WAW", ExecLayout.COMM_ARROW_COLOR );
	arrow.name = "waw";		
	addChild( arrow );
	sp_comm_arrows.push( arrow );
	
	g_timeline.addTween( ExecLayout.pathWipeBLtoTR(arrow, Anim.COMM_ARROW_WIPE) );
},
function():void {
	for each ( var s:Sprite in sp_comm_arrows ) { removeChild(s); }
	sp_comm_arrows = [];
},
(Anim.FADE) + (2*Anim.COMM_ARROW_WIPE));


// MOVE WAW ARROW DOWN TO SHOW INORDER COMMIT
var waw_mask:DisplayObject = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	var waw:Sprite = ExecLayout.findByName( sp_comm_arrows, "waw" );
	if ( waw.mask != null ) {
		waw_mask = waw.mask;
		waw.removeChild( waw_mask ); // TODO qqf waw_mask isn't a child of waw??
		waw.mask = null;
	}
	var commit:Sprite = ExecLayout.findByName( sp_commits, "commit2" );
	
	g_timeline.addTween( new Tween(waw,"y2",commit.y,
		Anim.Start(),Anim.StopIncr(Anim.COMM_ARROW_WIPE),Transitions.CUBIC_IN_AND_OUT) );
},
function():void {
	if ( waw_mask != null ) {
		var waw:Sprite = ExecLayout.findByName( sp_comm_arrows, "waw" );		
		waw.mask = waw_mask;
	}
},
Anim.COMM_ARROW_WIPE);

		SceneManager.noMoreScenes();
	} // end init()
	
}}