/*
    Copyright (c) 2009 by Joseph Devietti (devietti@cs.washington.edu).

    This file is part of my DMP ASPLOS 2009 Presentation (DAP).

    DAP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DAP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DAP.  If not, see <http://www.gnu.org/licenses/>.
*/ 

﻿package  {
	
	import com.dncompute.graphics.ArrowStyle;
	import com.dncompute.graphics.LineArrow;
	import flash.display.*;
	import flash.events.*;
	import flash.geom.Point;
	import flash.text.*;
	import flash.ui.*;
	import com.boostworthy.animation.management.*;
	import com.boostworthy.animation.easing.*;
	import com.boostworthy.animation.sequence.*;
	import com.boostworthy.animation.sequence.tweens.*;
	import com.boostworthy.animation.rendering.*;
	import com.boostworthy.core.Global;
	
	/**
	 * Animation for 10,000' overview of DMP
	 * @author Joe Devietti
	 */
	public class TenkView extends Sprite {
		
		// stage layout
		private const STAGE_WIDTH:int = 600;
		private const PANEL_WIDTH:Number = STAGE_WIDTH / 3.0;
		
		private const NUM_THREADS:int = 2;
		private const NUM_INSNS_PER_THREAD:int = 6;
		
		/** gap between threads within a panel */
		private const THREAD_GAP:Number = (PANEL_WIDTH - (2 * THREAD_WIDTH)) / 3;
		/** top of thread boxes */
		private const THREAD_TOP:int = 2;
		
		// a 3-panel panel, but centered across the whole stage
		private const PANEL0_START:Number = (STAGE_WIDTH - PANEL_WIDTH) / 2.0;
		// 3-panel layout
		private const PANEL1_START:Number = 0;
		private const PANEL2_START:Number = PANEL1_START + PANEL_WIDTH;
		private const PANEL3_START:Number = PANEL1_START + (2*PANEL_WIDTH);
		
		private const THREAD_WIDTH:int = 60;
		private const THREAD_HEIGHT:int = (NUM_INSNS_PER_THREAD*INSN_YOFFSET) + INSN_YGAP;
		
		private const INSN_WIDTH:int = THREAD_WIDTH-20;
		private const INSN_HEIGHT:int = 8;
		private const INSN_XOFFSET:int = (THREAD_WIDTH - INSN_WIDTH) / 2.0;
		private const INSN_YGAP:int = 5;
		private const INSN_YOFFSET:int = INSN_HEIGHT + INSN_YGAP;
		
		private const COMM_COLOR:uint = 0x0066CC;
		
		private function drawThread():Sprite {
			var t:Sprite = new Sprite();
			t.graphics.lineStyle();
			t.graphics.beginFill( 0xCCCCCC );
			t.graphics.drawRect( 0, 0, THREAD_WIDTH, THREAD_HEIGHT );
			t.graphics.endFill();
			t.filters = [ MyShapes.getDropShadow(0x0) ];
			return t;
		}
		private function drawHalfThread():Sprite {
			var t:Sprite = new Sprite();
			t.graphics.lineStyle();
			t.graphics.beginFill( 0xCCCCCC );
			t.graphics.drawRect( 0, 0, THREAD_WIDTH, THREAD_HEIGHT/2 );
			t.graphics.endFill();
			t.filters = [ MyShapes.getDropShadow(0x0) ];
			return t;
		}
		private function drawInsn():Sprite {
			var t:Sprite = new Sprite();
			t.graphics.lineStyle();
			t.graphics.beginFill( 0x0 );
			t.graphics.drawRect( 0, 0, INSN_WIDTH, INSN_HEIGHT );
			t.graphics.endFill();
			return t;
		}
		private function drawLabel(text:String):TextField {
			// create text
			var tf:TextField = new UnselectableTextField();
			tf.text = text;
			tf.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.align = TextFormatAlign.CENTER;
			format.font = "Tw Cen MT";
			format.color = 0x0;
			format.size = 30;
			format.bold = false;
			tf.setTextFormat( format );
			
			return tf;
		}
		private function drawEquiv():DisplayObject {
			// create text
			var tf:TextField = new UnselectableTextField();
			tf.text = "=";
			tf.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.align = TextFormatAlign.CENTER;
			format.font = "Tw Cen MT";
			format.color = COMM_COLOR;
			format.size = 50;
			format.bold = true;
			tf.setTextFormat( format );
			
			return tf;
		}
		
		public function TenkView() {
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private const g_timeline:Timeline = new Timeline( RenderMethod.ENTER_FRAME );
		
		/** offset is % below the top of the thread */
		private function vOffset(offset:Number):Number {
			return THREAD_TOP + (offset * THREAD_HEIGHT);
		}
		
		/** compute midpoints for each thread given the start of the panel */
		private function getThreadMidpoints( panel_start:int ):Array {
			var midpoints:Array = [];
			for ( var i:int = 0; i < NUM_THREADS; i++ ) {
				var t_start:Number = panel_start + ((i + 1) * THREAD_GAP) + (i * THREAD_WIDTH);
				midpoints[i] = t_start + (THREAD_WIDTH / 2.0);
			}
			return midpoints;
		}

		private function init(e:Event = null):void {
			removeEventListener(Event.ADDED_TO_STAGE, init);
			// entry point
			
			SceneManager.setup( g_timeline, stage );
			
			// local iteration vars
			var i:int = 0; var j:int = 0;
			var o:Object = null;
			var d:DisplayObject = null;
			var s:Sprite = null; var t:Sprite = null;
			
			var sp_orig_threads:Array = [];
			var sp_orig_insns:Array = [];
			
			// draw threads
			for ( i = 0; i < NUM_THREADS; i++ ) {
				t = drawThread();
				addChild( t );
				sp_orig_threads.push( t );

				t.x = PANEL0_START + ((i + 1) * THREAD_GAP) + (i * THREAD_WIDTH);
				t.y = THREAD_TOP;
				t.name = "t"+i;
				
				for ( j = 0; j < NUM_INSNS_PER_THREAD; j++ ) {
					s = drawInsn();
					s.name = ("i"+i)+j;
					addChild( s );
					sp_orig_insns.push( s );
					s.x = t.x + INSN_XOFFSET;
					s.y = t.y + INSN_YGAP + (INSN_YOFFSET * j);
					
				}
			}
			
			// label threads
			var sp_thread_labels:Array = [];
			for ( i=0; i < NUM_THREADS; i++ ) {
				var align:int = (i==0) ? MyShapes.LABEL_LEFT : MyShapes.LABEL_RIGHT ;
				var label:Sprite = MyShapes.Label( sp_orig_threads[i], "thread "+i, align );
				addChild( label );
				sp_thread_labels.push( label );
				
				if ( i == 0 ) {
					label.x = sp_orig_threads[i].x - MyShapes.LABEL_OFFSET;
				} else {
					label.x = sp_orig_threads[i].x + sp_orig_threads[i].width + MyShapes.LABEL_OFFSET;
				}
				label.y = sp_orig_threads[i].y;
			}


// LABEL INSNS
var sp_insn_labels:Array = [];
SceneManager.addScene(
function(start_frame:uint, end_frame:uint):void {
	// remove thread labels
	for each ( s in sp_thread_labels ) {
		g_timeline.addTween( ExecLayout.fadeToNoIncr(s,0.0) );
	}
	Anim.StopIncr(Anim.FADE);
	
	for each ( s in sp_orig_insns.concat(sp_orig_threads) ) {
		if ( s.name == "i04" ) { // label
			var label:Sprite = MyShapes.Label( s, "instruction", MyShapes.LABEL_LEFT, -3 );
			label.alpha = 0.0;
			addChild( label );
			sp_insn_labels.push( label );
			label.x = s.x - MyShapes.LABEL_OFFSET;
			label.y = s.y;
		} else if ( s.name == "i12" ) {
			label = MyShapes.Label( s, "instruction", MyShapes.LABEL_RIGHT, -3 );
			label.alpha = 0.0;
			addChild( label );
			sp_insn_labels.push( label );
			label.x = s.x + s.width + MyShapes.LABEL_OFFSET;
			label.y = s.y;
		} else { // dim
			g_timeline.addTween( ExecLayout.fadeToNoIncr(s,ExecLayout.ALPHA_DIM) );
		}
	}
	Anim.StopIncr(Anim.FADE);
	
	for each ( s in sp_insn_labels ) {
		g_timeline.addTween( ExecLayout.fadeToNoIncr(s,1.0) );
	}
	Anim.StopIncr(Anim.FADE);
	
	
},
function():void {
	for each ( var s:Sprite in sp_insn_labels ) { removeChild(s); }
	sp_insn_labels = [];
},
Anim.FADE*3);


// [src,dst] pairs
const COMM:Array = [ ["i00","i12"],["i02","i14"],["i04","i15"],["i10","i03"] ];

// SHOW COMMUNICATION
var sp_orig_comm_arrows:Array = [];
var sp_orig_label:TextField = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// dim insns labels
	for each ( s in sp_insn_labels ) {
		g_timeline.addTween( ExecLayout.fadeToNoIncr(s,0.0) );
	}
	for each ( s in sp_orig_insns.concat(sp_orig_threads) ) {
		g_timeline.addTween( ExecLayout.fadeToNoIncr(s,1.0) );
	}
	Anim.StopIncr(Anim.FADE);
	
	for each ( var arr:Array in COMM ) {
		var src:Sprite = ExecLayout.findByName( sp_orig_insns, arr[0] );
		var dst:Sprite = ExecLayout.findByName( sp_orig_insns, arr[1] );
		var startAtT0:Boolean = (arr[0].indexOf("i0") == 0);
		var arw:LineArrow;
		if ( startAtT0 ) { // start at t0
			arw = new LineArrow( src.x+src.width, src.y+(src.height/2), dst.x, dst.y+(dst.height/2), null, COMM_COLOR );
		} else { // start at t1
			arw = new LineArrow( src.x, src.y+(src.height/2), dst.x+dst.width, dst.y+(dst.height/2), null, COMM_COLOR );
		}
		arw.srcName = arr[0];
		arw.dstName = arr[1];
		addChild( arw );
		sp_orig_comm_arrows.push( arw );
		o = Wiper.wipe( arw, startAtT0 ? Wiper.WIPE_RIGHT : Wiper.WIPE_LEFT, true );
		g_timeline.addTween( new Tween(o.mask,"x",o.x,
			Anim.Start(),Anim.StopNoIncr(Anim.COMM_ARROW_WIPE),Transitions.LINEAR) );
	}
	Anim.StopIncr(Anim.COMM_ARROW_WIPE);
	
	// label
	sp_orig_label = drawLabel( "original\ncommunication" );
	sp_orig_label.alpha = 0.0;
	addChild( sp_orig_label );
	sp_orig_label.x = PANEL0_START + ((PANEL_WIDTH - sp_orig_label.width)/2);
	sp_orig_label.y = THREAD_HEIGHT;
	g_timeline.addTween( ExecLayout.fadeToIncr(sp_orig_label,1.0) );
},
function():void {
	for each ( var s:Sprite in sp_orig_comm_arrows ) { removeChild(s); }
	sp_orig_comm_arrows = [];
	removeChild( sp_orig_label );
	sp_orig_label = null;
},
(2*Anim.FADE) + (Anim.COMM_ARROW_WIPE));


// FADE AWAY NON-COMMUNICATING INSNS
SceneManager.addPowerpointClick();

SceneManager.addScene(
function(start_frame:uint, end_frame:uint):void {
	// build list of all comm insns' names
	var comm_insns:Array = [];
	for each ( var a:Array in COMM ) {
		comm_insns = comm_insns.concat(a);
	}
	
	for each ( s in sp_orig_insns ) {
		if ( comm_insns.indexOf(s.name) < 0 ) { // not a comm insn
			g_timeline.addTween( ExecLayout.fadeToNoIncr(s,ExecLayout.ALPHA_DIM) );
		}
	}
	Anim.StopIncr(Anim.FADE);
},
function():void {},
Anim.FADE);


SceneManager.addPowerpointClick();

// MOVE ORIG STUFF TO LEFT
var sp_ser_threads:Array = [];
var sp_ser_insns:Array = [];
var sp_ser_comm_arrows:Array = [];
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// remove masks from comm arrows, so they don't get masked out on move
	for each ( s in sp_orig_comm_arrows ) {
		var d:DisplayObject = s.mask;
		s.removeChild( d );
		s.mask = null;
	}
	
	// move everything to panel 1
	const XOFFSET:int = (PANEL0_START-PANEL1_START);
	for each ( d in sp_orig_insns.concat(sp_orig_threads,sp_orig_comm_arrows,sp_orig_label) ) {
		g_timeline.addTween( new Tween(d,"x",d.x-XOFFSET,
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	Anim.StopIncr(Anim.MOVE_TENK_THREADS);
},
function():void {},
Anim.MOVE_TENK_THREADS);


// SHOW SERIALIZATION
const SLIDE:Object = { // # half-threads to slide things down by, relative to current position
	i00:0,i01:0,i02:0,
	i03:1,i04:1,i05:1,
	i10:1,i11:1,i12:1,
	i13:2,i14:2,i15:2,
	t00:0,t01:1,t10:1,t11:2 };
var sp_ser_label:TextField = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// make a copy of everything (except label)...
	for each ( s in sp_orig_threads ) {
		for ( i = 0; i < 2; i++ ) {
			t = drawHalfThread();
			t.name = s.name+i;
			t.x = s.x; 
			t.y = s.y + (i * (THREAD_HEIGHT/2));
			addChild( t );
			sp_ser_threads.push( t );
		}
	}
	for each ( s in sp_orig_insns ) {
		t = drawInsn();
		t.x = s.x; t.y = s.y; t.alpha = s.alpha; t.name = s.name;
		addChild( t );
		sp_ser_insns.push( t );
	}
	for each ( var a:LineArrow in sp_orig_comm_arrows ) {
		var la:LineArrow = new LineArrow( a.x1, a.y1, a.x2, a.y2, null, COMM_COLOR );
		la.srcName = a.srcName; la.dstName = a.dstName;
		addChild( la );
		sp_ser_comm_arrows.push( la );
	}
	// ...and move it to panel 2
	for each ( s in sp_ser_threads.concat(sp_ser_insns,sp_ser_comm_arrows) ) {
		g_timeline.addTween( new Tween(s,"x",s.x+PANEL_WIDTH,
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	Anim.StopIncr(Anim.MOVE_TENK_THREADS);
	
	// slide stuff down
	const SLIDE_UNIT:Number = (THREAD_HEIGHT/2) + 4;
	for each ( s in sp_ser_threads.concat(sp_ser_insns) ) {
		g_timeline.addTween( new Tween(s,"y",s.y+(SLIDE[s.name] * SLIDE_UNIT),
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	for each ( la in sp_ser_comm_arrows ) {
		g_timeline.addTween( new Tween(la,"y1",la.y1+(SLIDE[la.srcName] * SLIDE_UNIT),
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
		g_timeline.addTween( new Tween(la,"y2",la.y2+(SLIDE[la.dstName] * SLIDE_UNIT),
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	Anim.StopIncr(Anim.MOVE_TENK_THREADS);
	
	// add label
	var sp_ser_label:TextField = drawLabel( "serialized" );
	sp_ser_label.alpha = 0.0;
	addChild( sp_ser_label );
	sp_ser_label.x = PANEL2_START + ((PANEL_WIDTH - sp_ser_label.width)/2);
	sp_ser_label.y = THREAD_HEIGHT*2 + 5; // HACK!!
	g_timeline.addTween( ExecLayout.fadeToIncr(sp_ser_label,1.0) );
},
function():void {
	for each ( var s:Sprite in sp_ser_threads.concat(sp_ser_insns,sp_ser_comm_arrows,sp_ser_label) ) { if (s!=null){removeChild(s)}; }
	sp_ser_threads = [];
	sp_ser_insns = [];
	sp_ser_comm_arrows = [];
	sp_ser_label = null;
},
(2*Anim.MOVE_TENK_THREADS) + (Anim.FADE));


SceneManager.addPowerpointClick();
SceneManager.addPowerpointClick();


// RECOVER PARALLELISM
var sp_par_threads:Array = [];
var sp_par_insns:Array = [];
var sp_par_comm_arrows:Array = [];
var sp_par_label:TextField = null;
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	// make a copy of everything...
	for each ( s in sp_ser_threads ) {
		t = drawHalfThread();
		t.name = s.name;
		t.x = s.x; t.y = s.y;
		addChild( t );
		sp_par_threads.push( t );
	}
	for each ( s in sp_ser_insns ) {
		t = drawInsn();
		t.x = s.x; t.y = s.y; t.alpha = s.alpha; t.name = s.name;
		addChild( t );
		sp_par_insns.push( t );
	}
	for each ( var a:LineArrow in sp_ser_comm_arrows ) {
		var la:LineArrow = new LineArrow( a.x1, a.y1, a.x2, a.y2, null, COMM_COLOR );
		la.srcName = a.srcName; la.dstName = a.dstName;
		addChild( la );
		sp_par_comm_arrows.push( la );
	}
	// ...and move it to panel 3
	for each ( s in sp_par_threads.concat(sp_par_insns,sp_par_comm_arrows) ) {
		g_timeline.addTween( new Tween(s,"x",s.x+PANEL_WIDTH,
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	Anim.StopIncr(Anim.MOVE_TENK_THREADS);
	
	// slide stuff up
	const SLIDE_UNIT:Number = (THREAD_HEIGHT/2) - 8;
	for each ( s in sp_par_threads.concat(sp_par_insns) ) {
		g_timeline.addTween( new Tween(s,"y",s.y-(SLIDE[s.name] * SLIDE_UNIT),
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	for each ( la in sp_par_comm_arrows ) {
		g_timeline.addTween( new Tween(la,"y1",la.y1-(SLIDE[la.srcName] * SLIDE_UNIT),
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
		g_timeline.addTween( new Tween(la,"y2",la.y2-(SLIDE[la.dstName] * SLIDE_UNIT),
			Anim.Start(),Anim.StopNoIncr(Anim.MOVE_TENK_THREADS),Transitions.CUBIC_IN_AND_OUT) );
	}
	Anim.StopIncr(Anim.MOVE_TENK_THREADS);
	
	// add label
	var sp_par_label:TextField = drawLabel( "parallelism\nrecovered" );
	sp_par_label.alpha = 0.0;
	addChild( sp_par_label );
	sp_par_label.x = PANEL3_START + ((PANEL_WIDTH - sp_par_label.width)/2);
	sp_par_label.y = THREAD_HEIGHT + 20; // HACK!
	g_timeline.addTween( ExecLayout.fadeToIncr(sp_par_label,1.0) );
},
function():void {
	for each ( var s:Sprite in sp_par_threads.concat(sp_par_insns,sp_par_comm_arrows,sp_par_label) ) { if(s!=null){removeChild(s)}; }
	sp_par_threads = [];
	sp_par_insns = [];
	sp_par_comm_arrows = [];
	sp_par_label = null;
},
Anim.FADE + (2*Anim.MOVE_TENK_THREADS));


// SHOW EQUIVALENCE OF ALL EXECUTIONS
var sp_equivalences:Array = [];
SceneManager.addScene( 
function(start_frame:uint, end_frame:uint):void {
	const y:int = 50;
	
	for each ( i in [PANEL2_START,PANEL3_START] ) {
		d = drawEquiv();
		d.alpha = 0.0;
		addChild( d );
		d.x = i - (d.width/2);
		d.y = y;
		g_timeline.addTween( ExecLayout.fadeToNoIncr(d,1.0) );	
	}
	
	/*
	d = drawEquiv();
	d.alpha = 0.0;
	addChild( d );
	d.x = PANEL3_START - (d.width/2);
	d.y = y;
	g_timeline.addTween( ExecLayout.fadeToNoIncr(d,1.0) );
	*/
	
	Anim.StopIncr(Anim.FADE);
},
function():void {
	for each ( var s:Sprite in sp_equivalences ) { removeChild(s); }
	sp_equivalences = [];
},
Anim.FADE);



			SceneManager.noMoreScenes();
		} // end init()
		
}}